/*--------------------------------------------------------------------------
| rtty.c - Simple 32-bit console TTY program to test serial-ports.
Copyright 1995,96,97 Comtrol Corp.
 6-04-97 BugFix - 7 databits reverted to using 8.
|--------------------------------------------------------------------------*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>

static int parse_cmdline(char *buf, char *port_name, DCB *dcb);
int SetToggleLow(HANDLE hCom, int on);

// Com stuff
HANDLE hCom;
char port_name[200];
char tmpstr[200];
DCB setup_dcb;  // our cmd line settings
unsigned char buf[600];
DWORD ModemStatus;
DWORD OldModemStatus;
char file_name[80];
char flow_control = 0;

int toggle = FALSE;       //true if rts toggle is was selected by user

/*--------------------------------------------------------------------------
| main -
|--------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
 unsigned char c;
 int stat, i;
 int get_info = 0;
 DCB our_dcb;
 DCB *dcb;
 char par_char, stop_char;
 DWORD ret_stat;
 DWORD cnt;
 COMSTAT ComStat;
 FILE *fp;
 int test_state = 0;
 int wr_size;
 DWORD byte_cnt = 0;

  file_name[0] = 0;
  port_name[0] = 0;
  memset(&setup_dcb, 0, sizeof(setup_dcb));

  setup_dcb.StopBits = 0;
  setup_dcb.BaudRate = 9600;
  setup_dcb.ByteSize = 8;
  setup_dcb.Parity = 0;

  printf("RTTY, 32-bit console mode terminal.\n");
  if (argc == 1)
  {
    printf("EXAMPLE>rtty COM5\n");
    printf("EXAMPLE>rtty COM3:1200\n");
    printf("EXAMPLE>rtty COM7:9600,8,n,1 \n");
    printf("EXAMPLE>rtty COM7:9600,8,n,1,r filesend.txt\n");
    printf("  parameters include: BAUD,databits,parity,stopbits,flowcontrol\n");
    printf("  where flowcontrol=X for xon,xoff, R=rts,cts,\n");
    printf("        4=rts toggle for 485 interface N=none\n");
#ifdef COMMENT_OUT
    return 1;
#endif

    printf("Enter COM port name to open(example:COM5):");
    scanf("%s", port_name);
    if (strlen(port_name) < 2)
    {
      printf("Bad Port Name\n");
      return 1;
    }
  }
  else
  {
    stat = parse_cmdline(argv[1], port_name, &setup_dcb);
  }

  if (argc == 3)  // have a file to send
  {
    strcpy(file_name, argv[2]);
    printf("sending %s file...\n", file_name);
  }

  tmpstr[0] = 0;
  strcpy(tmpstr,"\\\\.\\");
  strcat(tmpstr, port_name);
  hCom = CreateFile(tmpstr, GENERIC_READ | GENERIC_WRITE,
             0,                    // exclusive access
             NULL,                 // no security attrs
             OPEN_EXISTING,
             FILE_ATTRIBUTE_NORMAL,  // no overlapped io(keep it simple)
             NULL);
  if (hCom == INVALID_HANDLE_VALUE)
  {
    printf("Could not open %s\n",port_name);
    return 1;
  }
  else
  {
    printf("Opened %s\n",port_name);
  }

  dcb = &our_dcb;
  dcb->DCBlength = sizeof(DCB);
  stat = GetCommState(hCom, dcb);
  if (stat == TRUE) stat = 0;
  else stat = GetLastError();
  if (stat != 0)
  {
    printf("Error from GetCommState: %d\n",stat);
    return 1;
  }

  if (setup_dcb.BaudRate == 0)
    setup_dcb.BaudRate = 9600;

  if ((setup_dcb.ByteSize != 7) && (setup_dcb.ByteSize != 8))
    setup_dcb.ByteSize = 8;

  switch(setup_dcb.Parity)
  {
    case 0:  par_char = 'n'; break; // none
    case 1:  par_char = 'o'; break;
    case 2:  par_char = 'e'; break;
    case 3:  par_char = 'm'; break;
    case 4:  par_char = 's'; break;
  }

  switch (setup_dcb.StopBits)
  {
    case 0: stop_char = '1'; break;
    case 2: stop_char = '2'; break;
  }

  printf("Using %s:%d,%d,%c,%c... (Esc to Exit)\n",
      port_name,
      setup_dcb.BaudRate,
      setup_dcb.ByteSize,
      par_char,
      stop_char);

  dcb->StopBits = setup_dcb.StopBits;
  dcb->BaudRate = setup_dcb.BaudRate;
  dcb->ByteSize = setup_dcb.ByteSize;
  dcb->Parity   = setup_dcb.Parity;
  dcb->fDtrControl = DTR_CONTROL_ENABLE;   // turn on, leave on
  dcb->fRtsControl = RTS_CONTROL_ENABLE;   // turn on, leave on

  switch(flow_control)
  {
    case 'X': 
       printf("xon,xoff used\n");
       dcb->fDtrControl = DTR_CONTROL_ENABLE;   // turn on, leave on
       dcb->fRtsControl = RTS_CONTROL_ENABLE;   // turn on, leave on
       dcb->fOutxCtsFlow = FALSE;
       dcb->fOutX = TRUE;   // turn on xon,xoff flow
       dcb->fInX  = TRUE;   // turn on xon,xoff flow
    break;
    case 'R': 
       printf("rts,cts handshake used\n");
       dcb->fDtrControl = DTR_CONTROL_ENABLE;   // turn on, leave on
       dcb->fRtsControl = RTS_CONTROL_HANDSHAKE;   // flow control
       dcb->fOutxCtsFlow = TRUE;    // output flow ctrl enabled
       dcb->fOutX = FALSE;   // off
       dcb->fInX  = FALSE;   // off
    break;
    case 'N': 
       printf("no handshake used\n");
       dcb->fDtrControl = DTR_CONTROL_ENABLE;   // turn on, leave on
       dcb->fRtsControl = RTS_CONTROL_ENABLE;   // turn on, leave on
       dcb->fOutxCtsFlow = FALSE;
       dcb->fOutX = FALSE;   // off
       dcb->fInX  = FALSE;   // off
    break;
    case '4':
       dcb->fDtrControl = DTR_CONTROL_ENABLE;   // turn on, leave on
       dcb->fRtsControl = RTS_CONTROL_ENABLE;   // turn on, leave on
       dcb->fOutxCtsFlow = FALSE;
       dcb->fOutX = FALSE;   // off
       dcb->fInX  = FALSE;   // off
       printf("rts toggle mode set for 485\n");
       toggle = TRUE;
       break;
  }

  stat = SetCommState(hCom, dcb);
  if (stat != TRUE) {
    stat = GetLastError();
    printf("Error from SetCommState A3:%d",stat);
  }

  if (toggle == TRUE)
    SetToggleLow(hCom, 1);
  else
    SetToggleLow(hCom, 0);

//#define TRY_SOME_TIMEOUT_SETTINGS
#ifdef TRY_SOME_TIMEOUT_SETTINGS
{
COMMTIMEOUTS to;
   printf("try some timeouts\n");
   GetCommTimeouts(hCom, &to);
   to.WriteTotalTimeoutMultiplier = 10;   // ms per each character
   to.WriteTotalTimeoutConstant = 10;   // ms constant per write()
   to.ReadIntervalTimeout = 10;  //
   to.ReadTotalTimeoutMultiplier = 10;   // ms per each character
   to.ReadTotalTimeoutConstant = 10;   // ms constant per write()
   SetCommTimeouts(hCom, &to);
 }
#endif

//#define TRY_TXEMPTY_EVENTS
#ifdef TRY_TXEMPTY_EVENTS
 {
   OVERLAPPED  osEvent;     // for asynchronous operation
   DWORD Events;

   printf("starting overlap test\n");
   memset(&osEvent,0,sizeof(OVERLAPPED));
   osEvent.hEvent = CreateEvent(NULL,    // no security
                                TRUE,    // explicit reset needed
                                FALSE,   // initial event reset
                                NULL ) ; // no name
   if (NULL == osEvent.hEvent)
   {
     printf("error making event\n");
     exit(0);
   }

  SetCommMask(hCom, EV_TXEMPTY);
  printf("send data\n");
  memset(buf, 0x55, 150);
  WriteFile(hCom, buf, 150, &cnt, NULL);  // send out port
  stat = WaitCommEvent(hCom, &Events, NULL);
  //stat = WaitCommEvent(hCom, &Events, &osEvent);
  if (stat == FALSE)  // error
  {
    printf("error\n");
    stat = GetLastError();
    if (stat == ERROR_IO_PENDING)  // waiting for completion
      printf("io-pend\n");
  }
  else printf("got %xH, (TXEMPTY)\n", Events);

  printf("Test done.\n");
 }
 exit(0);
#endif

  // if file_name given at end of complete option list, then open file
  // send data out the port, and exit.
  if (file_name[0] != 0)
  {
    byte_cnt = 0;
    printf("Sending file: %s\n",file_name);
    fp = fopen(file_name, "rb");
    if (fp == NULL)
    {
      printf("bad filename\n");
    }

    printf("\n");
    wr_size = 100;
    while ((i = fread(buf,1,wr_size,fp)) > 0)
    {
      stat = WriteFile(hCom, buf, i, &cnt, NULL);  // send out port
      if (stat != TRUE)
      {
        printf("\x0d Error, %d from WriteFile\n",GetLastError());
        break;  // end while loop
      }
      if ((int)cnt != i)
      {
        printf("\x0d Error, %d sent, only %d complete\n",i, cnt);
        break;  // end while loop
      }

      byte_cnt += i;
      printf("\x0d %d bytes sent   ",byte_cnt);
    }
    printf("\n");

    if (setup_dcb.BaudRate <= 1200)
      Sleep(6000);  // give time for slow bauds to spool out

    printf("close\n");
    CloseHandle(hCom);
    exit(0);
  }

  // otherwise do normal tty stuff: display chars coming in, send out
  // any chars hit on the keyboard.

  stat = GetCommModemStatus(hCom, &ModemStatus);
  OldModemStatus = ModemStatus;
  printf("\nMODEM STATUS --->");
  if (ModemStatus & MS_CTS_ON) printf(" CTS ON");
  else printf(" CTS OFF");
  if (ModemStatus & MS_DSR_ON) printf(" DSR ON");
  else printf(" DSR OFF");
  if (ModemStatus & MS_RLSD_ON) printf(" CD ON\n");
  else printf(" CD OFF\n");


  c = 0;
  printf("Press ESC to exit(CAPS-LOCK must be off.)\n");
  while (c != 0x1b)
  {
    stat = GetCommModemStatus(hCom, &ModemStatus);
    if (stat == FALSE)
    {
      stat = GetLastError();
      printf("GetCommModemStatus Error %d (%xH)\n",stat,stat);
    }

    if (OldModemStatus != ModemStatus)
    {
      OldModemStatus = ModemStatus;
      printf("\nMODEM STATUS --->");
      if (ModemStatus & MS_CTS_ON) printf(" CTS ON");
      else printf(" CTS OFF");
      if (ModemStatus & MS_DSR_ON) printf(" DSR ON");
      else printf(" DSR OFF");
      if (ModemStatus & MS_RLSD_ON) printf(" CD ON\n");
      else printf(" CD OFF\n");
    }

    while (!kbhit())
    {
      stat = ClearCommError(hCom, &ret_stat, &ComStat);
      if (stat != TRUE) {
        stat = GetLastError();
        printf("Error ClearCommError = %d\n",stat);
      }
      else
      {
        if (ComStat.cbOutQue > 1)
          printf("Warning output buffer filling\n");

        if (ComStat.cbInQue > 0)
        {
           if (ComStat.cbInQue > 80) ComStat.cbInQue = 80; // limit
           stat = ReadFile(hCom, buf, ComStat.cbInQue, &cnt, NULL);
           if (stat != TRUE) {
             stat = GetLastError();
             printf("Error ReadFile = %d\n",stat);
           }
           else
           {
             buf[cnt] = 0;  // null terminate.
             printf("%s",buf);  // print rec. data to screen
           }
        }
      }
    }
    c = getch();    // grab keyboard char.
    if (c != 0x1b)
    {
      stat = WriteFile(hCom, &c, 1, &cnt, NULL);  // send out port
      if (stat != TRUE) {
        stat = GetLastError();
        printf("Error WriteFile = %d\n",stat);
      }
      else if (cnt != 1)
      {
        printf("Error WriteFile returned %d bytes\n",cnt);
      }
    }
  }

  EscapeCommFunction(hCom, CLRRTS);
  EscapeCommFunction(hCom, CLRDTR);
  CloseHandle(hCom);
 
  return(0);
}

/*--------------------------------------------------------------------
| parse_cmdline - Grab info in format "COM#:9600,8,n,1".
|---------------------------------------------------------------------*/
static int parse_cmdline(char *buf, char *port_name, DCB *dcb)
{
 int i;

  for (i=0; i<10; i++)
  {
    port_name[i] = buf[i];
    if ((buf[i] == ':') || (buf[i] == 0) || (buf[i] == ','))
    {
      port_name[i] = 0;
      break;
    }
  }

  if (buf[i] != ':') return 1;

  while ((*buf != 0) && (*buf != ':'))  // pass up COM#:
    ++buf;
  if (*buf == 0) return 0;  // end after COM#
  ++buf;
  if (!((*buf >= '0') && (*buf <= '9')))
  {
    printf("bad baudrate option\n");
    return 0;
  }
  dcb->BaudRate = atoi(buf);

  while ((*buf >= '0') && (*buf <= '9'))
   ++buf;
  if (*buf == 0) return 0;
  ++buf;
  if (!((*buf >= '0') && (*buf <= '9')))
  {
    printf("bad byte size option\n");
    return 0;
  }
  dcb->ByteSize = atoi(buf);

  while ((*buf >= '0') && (*buf <= '9'))
   ++buf;
  if (*buf == 0) return 0;
  ++buf;  // pass up ','

  switch(toupper(*buf++))
  {
    case 'N':  // none
      dcb->Parity = 0;
    break;
    case 'O':  // odd
      dcb->Parity = 1;
    break;
    case 'E':  // even
      dcb->Parity = 2;
    break;
    case 'M':  // mark
      dcb->Parity = 3;
    break;
    case 'S':  // space
      dcb->Parity = 4;
    break;
    default:
      printf("bad parity option\n");
      return 0;
    break;
  }

  while ((*buf != 0) && (*buf != ','))
    ++buf;
  if (*buf == 0) return 0;
  ++buf;
  switch(*buf)
  {
    case '1':
      dcb->StopBits = 0;
    break;
    case '2':
      dcb->StopBits = 2;
    break;
    default:
      printf("bad stopbit option\n");
      return 0;
  }
  ++buf;

  while ((*buf != 0) && (*buf != ','))
    ++buf;
  if (*buf == 0) return 0;
  ++buf;  // pass up ','

  switch(toupper(*buf++))
  {
    case 'X': flow_control = 'X'; break;
    case 'R': flow_control = 'R'; break;
    case 'N': flow_control = 'N'; break;
/////////////////new/////////////////
    case '4': flow_control = '4'; break;        
/////////////////////////////////////    
    default:
      printf("bad flowcontrol option\n");
      return 0;
  }

  //i = 0;
  //while (*buf != 0)
  //  file_name[i++] = *buf++;
  //file_name[i] = 0;

  return 0;
}

/*--------------------------------------------------------------------
| SetToggleLow - This function uses some custom ioctls which are
   only implemented in the RocketPort/VSlink product drivers.
   It is intended to demonstrate how to use a custom ioctl to
   turn on RS485 mode in the RS485 specific product.

   Override the port into RTS toggle mode on transmit.
   The override also sets the port to toggling RTS low on transmit(
   instead of the default high.)  This is what is needed for the RS485
   RocketPort product.
   This is intended for RS485 RocketPort product support, which
   requires RTS to toggle low on transmit.

  Params:
    hCom - Handle to port.
    on - set to 1 if we want 485 toggle override on, 0 to turn it off.
|---------------------------------------------------------------------*/
int SetToggleLow(HANDLE hCom, int on)
{
           // NT driver custom ioctl numbers for rocketport/vs
#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) )
#define METHOD_BUFFERED                 0
#define FILE_ANY_ACCESS                 0
#define FILE_DEVICE_SERIAL_PORT         0x0000001b
#define IOCTL_RCKT_LOOPBACK_ON \
      CTL_CODE(FILE_DEVICE_SERIAL_PORT, 0x805,METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_RCKT_LOOPBACK_OFF \
      CTL_CODE(FILE_DEVICE_SERIAL_PORT, 0x806,METHOD_BUFFERED,FILE_ANY_ACCESS)

#define IOCTL_RCKT_SET_TOGGLE_LOW \
      CTL_CODE(FILE_DEVICE_SERIAL_PORT, 0x807,METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_RCKT_CLEAR_TOGGLE_LOW \
      CTL_CODE(FILE_DEVICE_SERIAL_PORT, 0x808,METHOD_BUFFERED,FILE_ANY_ACCESS)

#define RTSTOGGLE_ON 15654L
#define RTSTOGGLE_OFF 15655L

  DWORD nBytes = 0;
  int stat;
  DWORD ioctl_id;
  DWORD PlatformID;
  OSVERSIONINFO os;  // about 150 bytes big

  os.dwOSVersionInfoSize = sizeof(os);
  if (FALSE == GetVersionEx(&os))  // nt or win95 or win32s
    PlatformID = VER_PLATFORM_WIN32_NT;
  else
    PlatformID = os.dwPlatformId;

    if (PlatformID == VER_PLATFORM_WIN32_NT)
    {
      if (on)
           ioctl_id = IOCTL_RCKT_SET_TOGGLE_LOW;
      else ioctl_id = IOCTL_RCKT_CLEAR_TOGGLE_LOW;

            // following only in RocketPort WinNT driver currently.
     stat = DeviceIoControl(hCom, ioctl_id,
                            (char *)NULL, 0, (char *)NULL, 0,  // out
                            &nBytes, 0);
    }
    else // win95 assume
    {
      if (on)
           ioctl_id = RTSTOGGLE_ON;
      else ioctl_id = RTSTOGGLE_OFF;

            // following only in RocketPort Win95 driver currently.
      stat = EscapeCommFunction(hCom, ioctl_id);  // low toggle on
    }
  return 0;
}
