/*-----------------------------------------------------------------------
| ttywin.cpp - handle tty text stuff to form tty style entry and display.
     Uses a circular queue structure to store the text.
Copyright 1997 Comtrol(TM) Corporation.
|-----------------------------------------------------------------------*/
/*** #include <windowsx.h> ***/
#define STRICT
#include "windows.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "ttywin.h"

int APIENTRY TTYtest(void);

// int __declspec(dllexport) TTYtest2(void)

#define ROW_HEIGHT 14

/*-------------------------------------------------------------------
| TTYwin_TTYwin - Constructor
|--------------------------------------------------------------------*/
int APIENTRY TTYtest(void)
{
  int i;
  TTYwin temp;

   i = 1;
   temp.cur_row = 0;
   temp.set_color(1);
   temp.set_size(1,2);
   temp.init(NULL);
   temp.mess_line("asdf", 1);
   temp.mess_str("asdf", 1);
   temp.mess_num(1);
   temp.clear_scr();

   return temp.cur_row;
}

/*-------------------------------------------------------------------
| TTYwin::TTYwin - Constructor
|--------------------------------------------------------------------*/
TTYwin::TTYwin()
{
  hwnd = 0;
  hfont = 0;

  screen_update_flag = 0;
  display_cur_row = 0;
  cur_row = 0;
  cur_col = 0;
  show_crlf = 0;
  scr_size_x = 80;
  scr_size_y = 20;
  text_color = RGB(0,0,0);  // black

  for (int i=0; i< TROWS;  i++)        /* scroll text */
  {
    text_buf[i][0] = 0;
    text_buf[i][TCOLS+2] = 0;
  }
}

/*-------------------------------------------------------------------
| TTYwin::~TTYwin - Destructor
|--------------------------------------------------------------------*/
TTYwin::~TTYwin()
{
}

/*------------------------------------------------------------------------
 TTYwin::init -
|------------------------------------------------------------------------*/
void TTYwin::init(HWND owner_hwnd)
{
  hwnd = owner_hwnd;
  CreateCaret(hwnd, (HBITMAP) NULL, 8, 8);
#if 0
  {
    char tmpstr[40];
    wsprintf(tmpstr, "%x", hwnd);
    MessageBox(GetFocus(),tmpstr,"",MB_OK);
  }
#endif
  caret_on = 0;
  hbrush_window = CreateSolidBrush(GetSysColor(COLOR_WINDOW));  // for painting window
}

/*------------------------------------------------------------------------
 TTYwin::set_color -
|------------------------------------------------------------------------*/
void TTYwin::set_color(int color_rgb)
{
  text_color = color_rgb;
}

/*------------------------------------------------------------------------
 TTYwin::show_caret -
|------------------------------------------------------------------------*/
void TTYwin::show_caret(int on)
{
 SIZE sz;
 int len;
 HDC hdc;
 int max_row;

  if (on)
  {
    max_row = ((scr_size_y - 4) / ROW_HEIGHT) - 1;
    //if (max_row < 1) return;

    caret_on = 1;
    CreateCaret(hwnd, (HBITMAP) NULL, 8, 8);

    len = lstrlen(text_buf[cur_row]);
    if (len != 0)
    {
      hdc = GetDC(hwnd);
      GetTextExtentPoint32(hdc, text_buf[cur_row], len, &sz);
      ReleaseDC(hwnd, hdc);
    }
    else
      sz.cx = 0;

    SetCaretPos(sz.cx+10,
                (max_row+1) * ROW_HEIGHT);
    ShowCaret(hwnd);
  }
  else
  {
    caret_on = 0;
    HideCaret(hwnd);
  }
}

/*------------------------------------------------------------------------
 TTYwin::set_size -
|------------------------------------------------------------------------*/
void TTYwin::set_size(int x, int y)
{
  scr_size_x = x;
  scr_size_y = y;
  update_screen(1);
}

/*------------------------------------------------------------------------
 TTYwin::update_screen - update anything which needs to be drawn on screen.
   Idea here is that the display data gets dumped into a circular
   buffer which we display to the screen(at a low priority).
|------------------------------------------------------------------------*/
void TTYwin::update_screen(int all_flag)
{
 int i,len;
 HDC hdc;
 int max_row;
 int write_y;
 int dis_row;
 char *txt;
 //LOGFONT lfont;
 HFONT new_hfont = NULL;
 //char tmparr[40];
 //TEXTMETRIC tm;

  if (!screen_update_flag)
  {
    if (!all_flag)
     return;  // no changes to screen.
  }
  screen_update_flag = 0;  // reset change detection.

  if (scr_size_y < (ROW_HEIGHT))
    return;
    
  max_row = ((scr_size_y - 4) / ROW_HEIGHT) - 1;
  if (max_row < 1) return;

  hdc = GetDC(hwnd);

  if (caret_on)
    HideCaret(hwnd);

#ifdef COMMENT_OUT
  if (hfont == NULL)
  {
    memset(&lfont, 0, sizeof(lfont));
    //strcpy(lfont.lfFaceName, "Terminal");

    lfont.lfHeight = 12;
    lfont.lfWidth = GetSystemMetrics(SM_CXSCREEN) / 100;
    lfont.lfCharSet = OEM_CHARSET; /* only one with hi-graphics chars */
    lfont.lfPitchAndFamily = FIXED_PITCH;

    hfont = CreateFontIndirect(&lfont);

    //sprintf(tmparr, "%s,X%d,Y%d", lfont.lfFaceName, lfont.lfWidth, lfont.lfHeight);
    //MessageBox(0, tmparr, "font CHAR", MB_OK | MB_ICONQUESTION);

    if (hfont != NULL)
      SelectObject(hdc,new_hfont);   // select new font


    GetTextMetrics(hdc, &tm);
    sprintf(tmparr, "aveX:%d, maxX:%d Y:%d",
                     tm.tmAveCharWidth,
                     tm.tmMaxCharWidth,
                     tm.tmHeight);

    MessageBox(0, tmparr, "Metrics CHAR", MB_OK | MB_ICONQUESTION);
  }
#endif

                             // implement almost normal TTY scrolling text
  if (display_cur_row != cur_row)  // force scroll
  {
    display_cur_row = cur_row;
    all_flag = 1;
  }
  dis_row = cur_row;

  for (i=max_row; i>=1; i--)
  {
    if ((text_buf[dis_row][TCOLS+2] != 0) || (all_flag))  // chg flag
    {
      txt = text_buf[dis_row];
      txt[TCOLS+2] = 0;  // reset chg flag
      len = lstrlen(txt);
      write_y = (i*ROW_HEIGHT+4);

      SelectObject(hdc,GetStockObject(NULL_PEN));
      //SelectObject(hdc,GetStockObject(WHITE_BRUSH));
      SelectObject(hdc,hbrush_window);
      if (i == max_row)
           Rectangle(hdc, 10, write_y, scr_size_x, write_y +ROW_HEIGHT+ROW_HEIGHT+1);
      else Rectangle(hdc, 10, write_y, scr_size_x, write_y +ROW_HEIGHT+1);
      if (len > 0)
      {
        //SetBkMode(hdc,OPAQUE);  // draw text with background
        //SetTextColor(hdc, text_color);
        TextOut(hdc, 10, write_y, txt, len);
      }
    }
    if (dis_row > 0) --dis_row;
    else dis_row = TROWS-1;
  }

  if (caret_on)
  {
    SIZE sz;
    txt = text_buf[cur_row];
    len = lstrlen(txt);
    if (len != 0)
      GetTextExtentPoint32(hdc, txt, len, &sz);
    else
      sz.cx = 0;

    SetCaretPos(sz.cx+10,
                (max_row+1) * ROW_HEIGHT);
    ShowCaret(hwnd);
  }

  ReleaseDC(hwnd, hdc);
}

/*------------------------------------------------------------------------
| TTYwin::mess_line -
|------------------------------------------------------------------------*/
void TTYwin::mess_line(char *str, int line_num)
{
 HDC hdc;

//  lstrcpy(text_buf, str);

  hdc = GetDC(hwnd);
  TextOut(hdc, 20, line_num*20, str, strlen(str));
  ReleaseDC(hwnd, hdc);
}

/*------------------------------------------------------------------------
| TTYwin::mess_num - send a number out to debug text window.
|------------------------------------------------------------------------*/
void TTYwin::mess_num(int num)
{
 char tmpstr[10];

  sprintf(tmpstr,":%d ",num);
  mess_str(tmpstr);
}

/*------------------------------------------------------------------------
| TTYwin::clear_scr - clear the screen
|------------------------------------------------------------------------*/
void TTYwin::clear_scr(void)
{
 int i;

  for (i=0; i<TROWS; i++) // for entire line
  {
    text_buf[i][0] = 0;  // null terminate
    text_buf[cur_row][TCOLS+2] = 1;  // chg flag
  }
  cur_row = 0;
  cur_col = 0;
  update_screen(1);
}

/*------------------------------------------------------------------------
| TTYwin::mess_str - send a string out to the text window.  Actually just
  dump it into our display circular queue buffer and set a flag to
  tell main poll routine we want to update the screen.  This avoids
  doing scrolling logic here, making it the display tasks problem,
  which is lower priority.
|------------------------------------------------------------------------*/
void TTYwin::mess_str(char *str, int len)
{
 int i;
 unsigned char c;
 char tmpstr[10];

 if (text_buf == NULL) return;
 if (str == NULL) return;

 if (len == 0)
   len = strlen(str);

 for (i=0; i<len; i++) // for entire line
 {
   c = (unsigned char) str[i];
   switch(c)
   {
     case 0xd:  // CR
       if (show_crlf)
         this->mess_str("<CR>");  // recurse
#ifdef ADD_LF
       if (cur_col != 0)
       {
         ++cur_row;
         if (cur_row >= TROWS)  // round queue
           cur_row = 0;
       }
#endif
       cur_col = 0;
       //  text_buf[cur_row][cur_col] = 0;
       text_buf[cur_row][TCOLS+2] = 1;  // chg flag
     break;

     case 0xa:  // LF
       ++cur_row;
       if (cur_row >= TROWS)  // round queue
         cur_row = 0;
       cur_col = 0;
       text_buf[cur_row][cur_col] = 0;
       text_buf[cur_row][TCOLS+2] = 1;  // chg flag

       if (show_crlf)
         this->mess_str("<LF>");  // recurse
     break;

     default:
       if ((c < 0x20) || (c > 0x80))
       {
         sprintf(tmpstr,"<%xH>",c);
         this->mess_str(tmpstr);  // recurse
         break;
       }
       text_buf[cur_row][cur_col] = c;
       text_buf[cur_row][cur_col+1] = 0;
       text_buf[cur_row][TCOLS+2] = 1;  // chg flag
       ++cur_col;
       if (cur_col >= TCOLS)  // end of line
       {
         cur_col = 0;
         ++cur_row;
         if (cur_row >= TROWS)  // round queue
            cur_row = 0;
         text_buf[cur_row][cur_col] = 0;
         text_buf[cur_row][TCOLS+2] = 1;  // chg flag
       }
     break;
   }
 }
 screen_update_flag = 1;  // need to update screen

}
